/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include <alloca.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <kstat.h>
#include <poll.h>
#include <sys/ioctl.h>
#include <sys/loadavg.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/swap.h>
#include <sys/systeminfo.h>
#include <sys/sysinfo.h>
#include <time.h>
#include <unistd.h>
#include <utmpx.h>


/* SUSv3 does not have PATH_MAX */
#ifndef PATH_MAX
#define PATH_MAX _XOPEN_PATH_MAX
#endif

#include "log.h"
#include "network.h"
#include "sysstat.h"

static kstat_ctl_t *KS = 0;
static sysstat_t Sysstat;

typedef struct cpudata_s {
	long long 
		user, kernel, idle, wait,
		intr, intrblk, xcalls,
		syscall, sysexec, sysfork, sysvfork,
		sysread, syswrite,
		pgswapin, pgswapout, scan,
		anonpgin, anonpgout, anonfree,
		phread, phwrite,
		inv_swtch, pswitch, migrate;
	hrtime_t
		user_ts, kernel_ts, idle_ts, wait_ts,
		timestamp;
	kstat_t 
		*cpustat, *cpusys;
	unsigned
		sys_iter, stat_iter;
} cpudata_t;


typedef struct class_chain_s {
	kstat_t *kstat;
	unsigned long long value[2];
	hrtime_t timestamp;
	unsigned iteration;
	struct class_chain_s *next;
} class_chain_t;

/* the number of cpus can change dynamically under Solaris */
static unsigned long
	Ocpu = 0,	/* number of online cpus */
	Maxid = 0,	/* highest id of a cpu */
	Iteration = 1,	/* used to make sure data is not update twice after a reconfig */
	Ncpus = 0;	/* number of CPUs available in the system (not necessaryly online) */

static cpudata_t
	**Cpus = 0;
kstat_t
	*Syspages = 0, *Ipstat = 0;

static class_chain_t
	*Err = 0, *Disks = 0, *Nfs = 0, *NfsServer = 0, *Net = 0;

/* cpu/sys percentages */
double UserP, KernelP, IdleP;
/* cpu/sys rates */
double IntrR, IntrBlkR, XcallsR, SyscallR, SysexecR, SysforkR,
       SysreadR, SyswriteR;
/* cpu/vm rates */
double InvSwtchR, PswitchR, MigrateR,
       PgswapinR, PgswapoutR, ScanR, AnonfreeR, AnonpginR, AnonpgoutR;
/* ip */
double IpInR, IpOutR, NetInR, NetOutR;
unsigned long IpInDiscards = 0, IpOutDiscards = 0;
/* unix/system_pages */
unsigned long Physmem, Pagesfree, Pageslocked, Pagestotal;
unsigned long Pgsz;
/* swap (non-kstat info) */
unsigned long SwapTotal, SwapFree;
/* device_errors */
unsigned long long Devices = 0, Storage;
unsigned long HardErrors = 0, SoftErrors = 0;
/* device I/O rates */
double DiskIR, DiskOR;
/* nfs I/O rates */
double NfsIR, NfsOR;
/* nfsd op rate */
double NfsdR;
/* boot time */
clock_t BootTime;

/* ip */
unsigned long IpInReceives, IpOutRequests, IpInErrs;
hrtime_t IpIn_ts, IpOut_ts;



static void
get_boottime(void)
{
	struct utmpx utmpx, *ent;
	utmpx.ut_type = BOOT_TIME;
	ent = getutxid(&utmpx);
	assert(ent);
	BootTime = ent->ut_tv.tv_sec;
}


static void
get_swapinfo(void)
{
	swaptbl_t *swaptbl;
	char *strtbl;
	int numswap,i;

	numswap = swapctl(SC_GETNSWP,0);
	assert(numswap != -1);
	numswap += 2;
	swaptbl = alloca(numswap * sizeof(swapent_t) + sizeof(struct swaptable));
	swaptbl->swt_n = numswap;
	strtbl = alloca(numswap * PATH_MAX);
	for (i = 0; i < numswap; ++i)
		swaptbl->swt_ent[i].ste_path = strtbl + (i * PATH_MAX);
	numswap = swapctl(SC_LIST,swaptbl);
	/* If you encounter the following assertion, then please
	 * recompile in 64 bit mode. It is caused by an interface
	 * limitation.
	 */
	assert(numswap != -1);
	SwapTotal = SwapFree = 0;
	for (i = 0; i < numswap; ++i) {
		SwapTotal += swaptbl->swt_ent[i].ste_pages;
		SwapFree += swaptbl->swt_ent[i].ste_free;
	}
	SwapTotal = ((long long)SwapTotal * Pgsz) >> 10;
	SwapFree = ((long long)SwapFree * Pgsz) >> 10;
}


static class_chain_t *
get_class_chain(const char *class, const char *module, const char *name, class_chain_t *c)
{
	kstat_t *k;
	class_chain_t *p = 0, *f = c;
	if (0 == c) {
		c = calloc(1,sizeof(class_chain_t));
		assert(c);
		f = c;
	}
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if ((class && strcmp(class,k->ks_class)) ||
		    (module && strcmp(module,k->ks_module)) ||
		    (name && strcmp(name,k->ks_name)))
			continue;
		if (0 == c) {
			c = calloc(1,sizeof(class_chain_t));
			/* p is always a valid pointer, because
			 * we allocate c before going into this loop
			 * and in every iteration if necessary
			 */
			p->next = c;
		} else {
			c->value[0] = c->value[1] = 0;
		}
		dbug("get_class_chain(): adding %s/%s/%s\n",k->ks_class,k->ks_module,k->ks_name);
		c->kstat = k;
		p = c;
		c = c->next;
	}
	while (c) {
		c->kstat = 0;
		c = c->next;
	}
	return f;
}


/* Solaris >=10 only */
static void
get_ip_ip(void)
{
	kstat_named_t *data;
	kid_t kid;

	kid = kstat_read(KS,Ipstat,0);
	assert(kid != -1);

	data = kstat_data_lookup(Ipstat,"inReceives");
	assert(data);
	assert(data->data_type == KSTAT_DATA_UINT32);
	IpInR = (double)(data->value.ui32 - IpInReceives) * 1E9 /
		(double)(Ipstat->ks_snaptime - IpIn_ts);
	IpIn_ts = Ipstat->ks_snaptime;
	IpInReceives = data->value.ui32;

	data = kstat_data_lookup(Ipstat,"outRequests");
	assert(data);
	assert(data->data_type == KSTAT_DATA_UINT32);
	IpOutR = (double)(data->value.ui32 - IpOutRequests) * 1E9/
		(double)(Ipstat->ks_snaptime - IpOut_ts);
	IpOut_ts = Ipstat->ks_snaptime;
	IpOutRequests = data->value.ui32;

	data = kstat_data_lookup(Ipstat,"inDiscards");
	assert(data);
	assert(data->data_type == KSTAT_DATA_UINT32);
	IpInDiscards = data->value.ui32;

	data = kstat_data_lookup(Ipstat,"outDiscards");
	assert(data);
	assert(data->data_type == KSTAT_DATA_UINT32);
	IpOutDiscards = data->value.ui32;
}


static int
get_device_errors(void)
{
	class_chain_t *c;

	HardErrors = 0;
	SoftErrors = 0;
	Storage = 0;
	Devices = 0;
	for (c = Err; c; c = c->next) {
		kstat_named_t *data;

		if (-1 == kstat_read(KS,c->kstat,0))
			return 1;
		assert(c->kstat->ks_type == KSTAT_TYPE_NAMED);
		data = kstat_data_lookup(c->kstat,"Size");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		if (data->value.ui64 == 0)
			continue;
		++Devices;
		Storage += data->value.ui64;
		data = kstat_data_lookup(c->kstat,"Hard Errors");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT32);
		HardErrors += data->value.ui32;
		data = kstat_data_lookup(c->kstat,"Soft Errors");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT32);
		SoftErrors += data->value.ui32;
	}
	Storage >>= 10;
	return 0;
}


static int
get_disk_io(void)
{
	class_chain_t *c;
	Devices = 0;
	for (c = Disks; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_io_t *data;
		double td;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert((s->ks_type == KSTAT_TYPE_IO)
			&& (s->ks_snaptime != c->timestamp));
		data = s->ks_data;
		td = 1E9 / (s->ks_snaptime - c->timestamp);
		if (c->value[0])
			DiskIR += (double)(data->nread - c->value[0]) * td;
		if (c->value[1])
			DiskOR += (double)(data->nwritten - c->value[1]) * td;
		c->value[0] = data->nread;
		c->value[1] = data->nwritten;
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	return 0;
}


static int
get_nfs_io(void)
{
	class_chain_t *c;
	for (c = Nfs; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_io_t *data;
		double td;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert(s->ks_type == KSTAT_TYPE_IO);
		data = s->ks_data;
		assert(s->ks_snaptime != c->timestamp);
		td = 1E9 / (s->ks_snaptime - c->timestamp);
		if (c->value[0])
			NfsIR += (double)(data->nread - c->value[0]) * td;
		assert(NfsIR < 1E10);
		if (c->value[1])
			NfsOR += (double)(data->nwritten - c->value[1]) * td;
		assert(NfsOR < 1E10);
		c->value[0] = data->nread;
		c->value[1] = data->nwritten;
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	for (c = NfsServer; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_named_t *data;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert(s->ks_type == KSTAT_TYPE_NAMED);
		data = kstat_data_lookup(s,"calls");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (c->value[0])
			NfsdR += (double)(data->value.ui64 - c->value[0]) * 1E9 /
				(s->ks_snaptime - c->timestamp);
		c->value[0] = data->value.ui64;
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	return 0;
}


static int
get_unix_systempages(void)
{
	kstat_named_t *data;

	if (kstat_read(KS,Syspages,0) == -1) {
		/* cannot be unconfigured/cannot fail */
		assert(0);
	}

	data = kstat_data_lookup(Syspages,"physmem");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Physmem = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Physmem = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pagesfree");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Pagesfree = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Pagesfree = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pageslocked");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Pageslocked = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Pageslocked = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pagestotal");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Pagestotal = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Pagestotal = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);
	return 0;
}


/* this is only for Solaris 10 and later */
static int
get_cpu_sys(void)
{
	int inst = 0, num = 0;

	/* not here, because for Solaris 9 this must be done in get_cpu_stat
	 * IdleP = UserP = KernelP = 0;
	 */
	for (inst = 0; inst < Maxid; ++inst) {
		kstat_named_t *data;
		cpudata_t *cpu;
		kstat_t *cpusys;
		unsigned long long user, kernel, idle;
		hrtime_t user_ts, kernel_ts, idle_ts;

		cpu = Cpus[inst];
		cpusys = cpu->cpusys;
		if (cpusys == 0)
			continue;
		if (kstat_read(KS,cpusys,0) == -1) {
			/* processor has been taken offline */
			Cpus[inst]->cpusys = 0;
			return 1;
		}
		++num;
		if (cpu->sys_iter == Iteration) {
			/* fields already up-to-date */
			continue;
		}
		cpu->sys_iter = Iteration;

		/* Every lookup here may fail, but the whole set of metrics
		 * should be committed as a transaction, because otherwise
		 * we won't be able to handle dynamic reconfiguration
		 * correctly. So first do all lookups and then write the
		 * results back. Every lookup modifies the ks_snaptime, so
		 * we need to have seperate timestamps.
		 */
		data = kstat_data_lookup(cpusys,"cpu_nsec_user");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		user = data->value.ui64;
		user_ts = cpusys->ks_snaptime;
		data = kstat_data_lookup(cpusys,"cpu_nsec_kernel");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		kernel = data->value.ui64;
		kernel_ts = cpusys->ks_snaptime;
		data = kstat_data_lookup(cpusys,"cpu_nsec_idle");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		idle = data->value.ui64;
		idle_ts = cpusys->ks_snaptime;

		UserP += (double)(user - cpu->user)
			/ (double)(user_ts - cpu->user_ts)
			/ Ocpu * 100;
		cpu->user = user;
		cpu->user_ts = user_ts;
		KernelP += (double)(kernel - cpu->kernel)
			/ (double)(kernel_ts - cpu->kernel_ts)
			/ Ocpu * 100;
		cpu->kernel = kernel;
		cpu->kernel_ts = kernel_ts;
		IdleP += (double)(idle - cpu->idle)
			/ (double)(idle_ts - cpu->idle_ts)
			/ Ocpu * 100;
		cpu->idle = idle;
		cpu->idle_ts = idle_ts;
	}
	if (num >= Ocpu)
		return 0;
	return 1;
}


static int
get_cpu_stat(void)
{
	unsigned inst = 0, num = 0;

	for (inst = 0; inst < Maxid; ++inst) {
		cpudata_t *cpu = Cpus[inst];
		kstat_t *cpustat = cpu->cpustat;
		cpu_stat_t *stat;
		double td;

		if (cpustat == 0)
			continue;
		if (-1 == kstat_read(KS,cpustat,0)) {
			/* processor was taken offline */
			return 1;
		}

		++num;
		if (cpu->stat_iter == Iteration) {
			/* this info was updated before reconfiguring */
			continue;
		}
		stat = cpustat->ks_data;
		cpu->stat_iter = Iteration;
		
		td = 1 / ((cpustat->ks_snaptime - cpu->timestamp) * 1E-9);
		IntrR += (double)(stat->cpu_sysinfo.intr - cpu->intr) * td;
		cpu->intr = stat->cpu_sysinfo.intr;
		IntrBlkR += (double)(stat->cpu_sysinfo.intrblk - cpu->intrblk) * td;
		cpu->intrblk = stat->cpu_sysinfo.intrblk;
		XcallsR += (double)(stat->cpu_sysinfo.xcalls - cpu->xcalls) * td;
		cpu->xcalls = stat->cpu_sysinfo.xcalls;
		PswitchR += (double)(stat->cpu_sysinfo.pswitch - cpu->pswitch) * td;
		cpu->pswitch = stat->cpu_sysinfo.pswitch;
		MigrateR += (double)(stat->cpu_sysinfo.cpumigrate - cpu->migrate) * td;
		cpu->migrate = stat->cpu_sysinfo.cpumigrate;
		InvSwtchR += (double)(stat->cpu_sysinfo.inv_swtch - cpu->inv_swtch) * td;
		cpu->inv_swtch = stat->cpu_sysinfo.inv_swtch;
		SyscallR += (double)(stat->cpu_sysinfo.syscall - cpu->syscall) * td;
		cpu->syscall = stat->cpu_sysinfo.syscall;
		SysforkR += (double)(stat->cpu_sysinfo.sysfork - cpu->sysfork) * td;
		cpu->sysfork = stat->cpu_sysinfo.sysfork;
		SysforkR += (double)(stat->cpu_sysinfo.sysvfork - cpu->sysvfork) * td;
		cpu->sysvfork = stat->cpu_sysinfo.sysvfork;
		SysexecR += (double)(stat->cpu_sysinfo.sysexec - cpu->sysexec) * td;
		cpu->sysexec = stat->cpu_sysinfo.sysexec;
		SysreadR += (double)(stat->cpu_sysinfo.sysread - cpu->sysread) * td;
		cpu->sysread = stat->cpu_sysinfo.sysread;
		SyswriteR += (double)(stat->cpu_sysinfo.syswrite - cpu->syswrite) * td;
		cpu->syswrite = stat->cpu_sysinfo.syswrite;
		AnonpginR += (double)(stat->cpu_vminfo.anonpgin - cpu->anonpgin) * td;
		cpu->anonpgin = stat->cpu_vminfo.anonpgin;
		AnonpgoutR += (double)(stat->cpu_vminfo.anonpgout - cpu->anonpgout) * td;
		cpu->anonpgout = stat->cpu_vminfo.anonpgout;
		PgswapinR += (double)(stat->cpu_vminfo.pgswapin - cpu->pgswapin) * td;
		cpu->pgswapin = stat->cpu_vminfo.pgswapin;
		PgswapoutR += (double)(stat->cpu_vminfo.pgswapout - cpu->pgswapout) * td;
		cpu->pgswapout = stat->cpu_vminfo.pgswapout;
		ScanR += (double)(stat->cpu_vminfo.scan - cpu->scan) * td;
		cpu->scan = stat->cpu_vminfo.scan;
		AnonfreeR += (double)(stat->cpu_vminfo.anonfree - cpu->anonfree) * td;
		cpu->anonfree = stat->cpu_vminfo.anonfree;
		cpu->timestamp = cpustat->ks_snaptime;

		if (Cpus[inst]->cpusys == 0) {
			int total = stat->cpu_sysinfo.cpu[CPU_USER]
				+ stat->cpu_sysinfo.cpu[CPU_KERNEL]
				+ stat->cpu_sysinfo.cpu[CPU_IDLE]
				+ stat->cpu_sysinfo.cpu[CPU_WAIT]
				- cpu->user
				- cpu->kernel
				- cpu->idle
				- cpu->wait;
			UserP += (double)(stat->cpu_sysinfo.cpu[CPU_USER] - cpu->user)
				/ (double)(total)
				/ Ocpu * 100;
			cpu->user = stat->cpu_sysinfo.cpu[CPU_USER];
			KernelP += (double)(stat->cpu_sysinfo.cpu[CPU_KERNEL] - cpu->kernel)
				/ (double)(total)
				/ Ocpu * 100;
			cpu->kernel = stat->cpu_sysinfo.cpu[CPU_KERNEL];
			IdleP += (double)(stat->cpu_sysinfo.cpu[CPU_IDLE] - cpu->idle)
				/ (double)(total)
				/ Ocpu * 100;
			cpu->idle = stat->cpu_sysinfo.cpu[CPU_IDLE];
			IdleP += (double)(stat->cpu_sysinfo.cpu[CPU_WAIT] - cpu->wait)
				/ (double)(total)
				/ Ocpu * 100;
			cpu->wait = stat->cpu_sysinfo.cpu[CPU_WAIT];
		}
	}
	if (num == Ocpu)
		return 0;
	assert(num < Ocpu);
	return 1;
}


static int
get_net_stats(void)
{
	class_chain_t *n = Net;
	NetInR = NetOutR = 0;
	while (n) {
		kstat_named_t *data;
		if (kstat_read(KS,n->kstat,0) == -1) {
			/* network interface has been taken offline */
			n->kstat = 0;
			return 1;
		}
		assert(n->kstat->ks_type == KSTAT_TYPE_NAMED);
		data = kstat_data_lookup(n->kstat,"ipackets64");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (n->value[0])
			NetInR += (double)(data->value.ui64 - n->value[0]) / 
				(double)(n->kstat->ks_snaptime - n->timestamp) * 1E9;
		n->value[0] = data->value.ui64;
		data = kstat_data_lookup(n->kstat,"opackets64");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (n->value[1])
			NetOutR += (double)(data->value.ui64 - n->value[1]) / 
				(double)(n->kstat->ks_snaptime - n->timestamp) * 1E9;
		n->value[1] = data->value.ui64;
		n->timestamp = n->kstat->ks_snaptime;
		n = n->next;
	}
	return 0;
}


static void
setup_nets(void)
{
	class_chain_t *n, *p = 0;
	Net = get_class_chain("net",0,0,Net);
	n = Net;
	while (n) {
		/* The loopback interface "lo0" doesn't seem to have
		 * a ipackets64 field. So we use this field to be able
		 * to ignore it...
		if ((0 == strcmp(n->kstat->ks_name,"zero_copy")) ||
		    (0 == strcmp(n->kstat->ks_name,"lo0")) ||
		 */
		if ((-1 == kstat_read(KS,n->kstat,0)) ||
		    (0 == kstat_data_lookup(n->kstat,"ipackets64"))) {
			class_chain_t *l = n;
			if (p)
				p->next = n->next;
			else
				Net = n->next;
			n = n->next;
			free(l);
		} else
			n = n->next;
	}
}


static void
setup_cpus(void)
{
	unsigned num = 0, maxid = Maxid;
	int i;
	kstat_t *k;

#ifdef _SC_CPUID_MAX
	Maxid = sysconf(_SC_CPUID_MAX);
#else
	/* the maximum CPUID cannot be asked on Solaris 8 systems.
	 * Therefore, the value of NCPU in sys/param.h is used as 
	 * maximum CPUID, which should work for almost all cases.
	 * NCPU is a kernel limit for number of CPUs, _not_ maximum ID.
	 * So this is _not_ perfect, but good enough.
	 */
	Maxid = 56;
#endif
	
	/* number of cpu change during runtime */
	while (maxid > Maxid) {
		free(Cpus[maxid--]);
	}
	Cpus = realloc(Cpus,sizeof(cpudata_t *) * (Maxid + 1));
	assert(Cpus);
	while (maxid <= Maxid) {
		Cpus[maxid] = calloc(1,sizeof(cpudata_t));
		assert(Cpus[maxid]);
		++maxid;
	}
	assert(Ncpus != 0);
	for (i = 0; i <= Maxid; ++i) {
		Cpus[i]->cpustat = 0;
		Cpus[i]->cpusys = 0;
	}
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if (0 == strcmp(k->ks_module,"cpu_stat")) {
			assert(k->ks_type == KSTAT_TYPE_RAW);
			assert(k->ks_instance <= maxid);
			Cpus[k->ks_instance]->cpustat = k;
			dbug("found CPU with id %d\n",k->ks_instance);
			++num;
		} else if ((0 == strcmp(k->ks_module,"cpu"))
				&& (0 == strcmp(k->ks_name,"sys"))) {
			assert(k->ks_instance <= maxid);
			Cpus[k->ks_instance]->cpusys= k;
			assert(k->ks_type == KSTAT_TYPE_NAMED);
		}
	}
	assert(num <= Ncpus);
	if (Ocpu != num) {
		/* Try to reduce the transient effect of onlining of CPUs on
		 * metrics calculation.
		 */
		UserP /= Ocpu;
		UserP *= num;
		KernelP /= Ocpu;
		KernelP *= num;
		IdleP /= Ocpu;
		IdleP *= num;
		Ocpu = num;
		dbug("Ocpu = %d\n",Ocpu);
	}
}


static int
reconfigure(void)
{
	static kstat_t *sysmisc = 0;
	static kstat_named_t *data = 0;
	int upd;
	kid_t kid;
	class_chain_t *c;

	upd = kstat_chain_update(KS);
	if ((0 == upd) && Ncpus)
		return 1;
	if (-1 == upd)
		error("Kstat update failed: %s\n",strerror(errno));
	if (sysmisc == 0) {
		sysmisc = kstat_lookup(KS,"unix",0,"system_misc");
		assert(sysmisc && (sysmisc->ks_type == KSTAT_TYPE_NAMED));
	}
	kid = kstat_read(KS,sysmisc,0);
	assert(kid != -1);
	data = kstat_data_lookup(sysmisc,"ncpus");
	assert(data && data->data_type == KSTAT_DATA_UINT32);
	Ncpus = data->value.ui32;
	setup_cpus();
	Err = get_class_chain("device_error",0,0,Err);
	Disks = get_class_chain("disk","sd",0,Disks);
	assert(Disks);
	c = Disks;
	while (c->kstat && c->next)
		c = c->next;
	if (c->kstat) {
		if (c->next == 0)
			c->next = calloc(1,sizeof(class_chain_t));
		c = c->next;
	}
	assert(c->kstat == 0);
	(void) get_class_chain("disk","cmdk",0,c);
	Nfs = get_class_chain("nfs","nfs",0,Nfs);
	NfsServer = get_class_chain("misc","nfs","nfs_server",NfsServer);
	return 1;
}


void
get_local_data(sysstat_t *s)
{
	IntrR = IntrBlkR = XcallsR = 0;
	SyscallR = SysexecR = SysforkR = 0;
	SysreadR = SyswriteR = 0;
	InvSwtchR = PswitchR = MigrateR = 0;
	IdleP = UserP = KernelP = 0;
	DiskIR = DiskOR = 0;
	NfsdR = NfsIR = NfsOR = 0;
	ScanR = PgswapinR = PgswapoutR = 0;
	AnonfreeR = AnonpginR = AnonpgoutR = 0;
	while (reconfigure()) {
		clock_t now;

		if (get_cpu_stat())
			continue;
		if (get_disk_io())
			continue;
		if (update_fs_stats())
			continue;
		if (get_nfs_io())
			continue;
		if (get_unix_systempages())
			continue;
		if (Ipstat) {
			/* Solaris 10 extras */
			if (get_cpu_sys())
				continue;
			get_ip_ip();
		} else if (get_net_stats())
			continue;
		get_device_errors();
		get_swapinfo();
		++Iteration;
		if (s == 0)
			return;
		now = time(0);
		s->TS = now;
		s->BootTime = BootTime;
		getloadavg(s->loadavg,3);
		s->Ncpus = Ncpus;
		s->Ocpus = Ocpu;
		s->Devices = Devices;
		s->Storage = Storage;
		s->Zsize = Zsize;
		s->Zfree = Zfree;
		s->Zpools = Zpools;
		s->HardErrors = HardErrors;
		s->SoftErrors = SoftErrors;
		s->Physmem = Physmem;
		s->Pagesfree = Pagesfree;
		s->Pagestotal = Pagestotal;
		s->Pageslocked = Pageslocked;
		s->SwapTotal = SwapTotal;
		s->SwapFree = SwapFree;
		s->ScanR = ScanR;
		s->AnonpginR = AnonpginR;
		s->AnonpgoutR = AnonpgoutR;
		s->AnonfreeR = AnonfreeR;
		s->PgswapinR = PgswapinR;
		s->PgswapoutR = PgswapoutR;
		s->IntrR = IntrR;
		s->IntrBlkR = IntrBlkR;
		s->XcallsR = XcallsR;
		s->SyscallR = SyscallR;
		s->SysexecR = SysexecR;
		s->SysforkR = SysforkR;
		s->SysreadR = SysreadR;
		s->SyswriteR = SyswriteR;
		s->InvSwtchR = InvSwtchR;
		s->PswitchR = PswitchR;
		s->MigrateR = MigrateR;
		s->IdleP = IdleP;
		s->UserP = UserP;
		s->KernelP = KernelP;
		s->DiskIR = DiskIR/1024;
		s->DiskOR = DiskOR/1024;
		s->NfsIR = NfsIR/1024;
		s->NfsOR = NfsOR/1024;
		s->NfsdR = NfsdR;
		s->FsFree = FsFree;
		s->IpInR = IpInR;
		s->IpOutR = IpOutR;
		s->NetInR = NetInR;
		s->NetOutR = NetOutR;
		s->IpInDiscards = IpInDiscards;
		s->IpOutDiscards = IpOutDiscards;
		s->IpInErrs = IpInErrs;
		s->fs = get_fs_stats();
		s->us = get_procinfo();
		s->displayed = 0;
		return;
	}
}


void
init_localstat(sysstat_t *s)
{
	int ret;
	char *next;
	struct rlimit rlimit;

	if (s == 0)
		s = &Sysstat;
	if (-1 == getrlimit(RLIMIT_NOFILE,&rlimit))
		info("could not get resource limit: %s\n",strerror(errno));
	rlimit.rlim_cur = rlimit.rlim_max;
	if (-1 == setrlimit(RLIMIT_NOFILE,&rlimit))
		warn("could not reset no of files limit: %s\n",strerror(errno));
	get_boottime();
	KS = kstat_open();
	if (KS == 0)
		error("cannot access kernel statistics: %s\n",strerror(errno));
	Syspages = kstat_lookup(KS,"unix",0,"system_pages");
	assert(Syspages && (Syspages->ks_type == KSTAT_TYPE_NAMED));
	Ipstat = kstat_lookup(KS,"ip",0,"ip");
	if (Ipstat == 0) {
		/* before Solaris 10 */
		IpInR = IpOutR = -1;
		setup_nets();
	} else 
		assert(Ipstat->ks_type == KSTAT_TYPE_NAMED);
	Pgsz = getpagesize();
	ret = gethostname(s->Hostinfo,sizeof(s->Hostinfo));
	assert(ret == 0);
	if ((next = strchr(s->Hostinfo,'.')) == 0)
		next = s->Hostinfo + strlen(s->Hostinfo);
	*next = ' ';
	++next;
	sysinfo(SI_SYSNAME,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	next += strlen(next);
	*next = ' ';
	++next;
	sysinfo(SI_RELEASE,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	next += strlen(next);
	*next = ' ';
	++next;
	sysinfo(SI_PLATFORM,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	reconfigure();
	/* getdata once then sleep 0.3sec to get reasonable start values */
	get_local_data(0);
	(void) usleep(300000);	
}


void
daemon_sighandler(int s)
{
	assert(s == SIGTERM || s == SIGQUIT || s == SIGINT);
	(void) unlink("/tmp/sysstat.pid");
	exit(EXIT_SUCCESS);
}


void
rundaemon(int interval)
{
	log2syslog();
	if (-1 == chdir("/var/tmp"))
		error("Could not change directory to /var/tmp: %s\n",strerror(errno));
	(void) signal(SIGTERM,daemon_sighandler);
	(void) signal(SIGQUIT,daemon_sighandler);
	(void) signal(SIGINT,daemon_sighandler);
	init_zfs();
	init_localstat(&Sysstat);
	for (;;) {
		get_local_data(&Sysstat);
		send_to_network(&Sysstat);
		(void) sleep(interval);
	}
}

// vim:tw=0
